#!/usr/bin/env python3

##
# @file netcdf2yac.py
# @brief reading a netcdf file and send its content via yac to the other component
#
# This example reads a NetCDF file and registers the
# defined grid to yac. All data fields with correct dimensions are
# registered to yac as well. In the time loop the fields are read and send to
# the other component via the `Field.put` function.
#
# @copyright Copyright  (C)  2022 DKRZ, MPI-M
#
# @author Nils-Arne Dreier <dreier@dkrz.de>
#

#
# Keywords:
# Maintainer: Nils-Arne Dreier <dreier@dkrz.de>
# URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
#
# This file is part of YAC.
#
# Redistribution and use in source and binary forms, with or without
# modification, are  permitted provided that the following conditions are
# met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# Neither the name of the DKRZ GmbH nor the names of its contributors
# may be used to endorse or promote products derived from this software
# without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import argparse

parser = argparse.ArgumentParser(description=
"""Read variables NetCDF File with structured or unstructured grid and put it into YAC!
This is an example program and not thought to be working for generic NetCDF files.
Please look at the code to determine which format (variables) are expected.
""")
parser.add_argument(
    "filename", type=str, help="NetCDF file"
)
args = parser.parse_args()


from yac import *
import xarray as xr
import numpy as np

yac = YAC()
def_calendar(Calendar.PROLEPTIC_GREGORIAN)

dataset = xr.open_dataset(args.filename, use_cftime=True)

time = [np.datetime64(str(t)) for t in dataset.time.values]
timestep = time[1]-time[0]
yac.def_datetime(str(time[0])[:23], str(time[-1])[:23])

comp = yac.def_comp(f"NetCDF")
deg2rad = np.pi/180
if {"lat", "lon"} <= dataset.dims.keys():
    lon = deg2rad*dataset["lon"]
    lat = deg2rad*dataset["lat"]
    dims = ["lat", "lon"]
    grid = Reg2dGrid(f"netcdf_grid", lon, lat)
    points = grid.def_points(Location.CORNER, lon, lat)
elif {"vertices", "i"} <= dataset.dims.keys():
    vertices_per_cell = dataset.dims["vertices"]
    dims = ["i"]
    vtx = np.array([deg2rad*dataset["longitude_bnds"].values.ravel(), deg2rad*dataset["latitude_bnds"].values.ravel()])
    vtx, cellidx = np.unique(vtx, return_inverse=True, axis=1)
    grid = UnstructuredGrid("netcdf_grid", np.ones(dataset.dims["i"])*vertices_per_cell,
                            vtx[0,:], vtx[1,:], cellidx)
    points = grid.def_points(Location.CELL, deg2rad*dataset["longitude"], deg2rad*dataset["latitude"])
else:
    raise RuntimeError("Cant detect grid!")

fields = {}
for name, var in dataset.items():
    # check shape:
    if "time" in var.dims and set(dims) <= set(var.dims):
        print("define field: ", name)
        fields[name] = Field.create(name, comp, points, 1, str(timestep // np.timedelta64(1, 'ms')),
                                    TimeUnit.MILLISECOND)
    else:
        print("ignoring field: ", name)

yac.enddef()
print("starttime: ", yac.start_datetime)
print("endtime: ", yac.end_datetime)
for i, t in enumerate(time):
    print(f"call yac_put for timestep {t}")
    for name, field in fields.items():
        if field.role is ExchangeType.SOURCE:
            field.put(dataset[name][{'time': i}].values.reshape((1,-1)))
