#!/usr/bin/env python3

##
# @file plot.py
# @brief receiving a field on a structured grid and plot it using matplotlib
#
# \example plot.py
# This example registers given fields on a structured grid using the
# `Reg2dGrid` function. In the time loop it receives the data using
# `Field.get` function and generates a png image using matplotlib.
#
# @copyright Copyright  (C)  2022 DKRZ, MPI-M
#
# @author Nils-Arne Dreier <dreier@dkrz.de>
#

#
# Keywords:
# Maintainer: Nils-Arne Dreier <dreier@dkrz.de>
# URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
#
# This file is part of YAC.
#
# Redistribution and use in source and binary forms, with or without
# modification, are  permitted provided that the following conditions are
# met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# Neither the name of the DKRZ GmbH nor the names of its contributors
# may be used to endorse or promote products derived from this software
# without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


import os
from yac import *
import numpy as np
import matplotlib.pyplot as plt
import isodate
import argparse

parser = argparse.ArgumentParser(description="Plot variables in an YAC environment on a regular grid",
                                 formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('-v','--variables', nargs='+',
                    help='List of variables to plot', required=True)
parser.add_argument(
    "--res_lat",
    type=int,
    default=50,
    help="Resolution for latitude",
)
parser.add_argument(
    "--res_lon",
    type=int,
    default=100,
    help="Resolution for longitude",
)
parser.add_argument(
    "--min_lat",
    type=float,
    default=-0.5*np.pi,
    help="lower boundary for latitude",
)
parser.add_argument(
    "--min_lon",
    type=float,
    default=0,
    help="lower boundary for longitude",
)
parser.add_argument(
    "--max_lat",
    type=float,
    default=0.5*np.pi,
    help="upper boundary for latitude",
)
parser.add_argument(
    "--max_lon",
    type=float,
    default=2*np.pi,
    help="upper boundary for longitude",
)

args = parser.parse_args()

yac = YAC()

def_calendar(Calendar.PROLEPTIC_GREGORIAN)

comp = yac.def_comp(f"plot")

x = np.linspace(args.min_lon,args.max_lon,args.res_lon)
y = np.linspace(args.min_lat,args.max_lat,args.res_lat)
grid = Reg2dGrid(f"plot_grid", x, y)
points = grid.def_points(Location.CORNER, x, y)

yac.sync_def()

nnn = InterpolationStack()
nnn.add_nnn(NNNReductionType.AVG, 1, 1.)

fields = []
for comp_name in yac.component_names:
    for grid_name in yac.grid_names:
        # check for variables to plot
        for varname in args.variables:
            field_names = yac.get_field_names(comp_name, grid_name)
            if varname in field_names:
                my_field_name = f"{comp_name}/{grid_name}/{varname}"
                collection_size = yac.get_field_collection_size(comp_name, grid_name, varname)
                timestep = yac.get_field_timestep(comp_name, grid_name, varname)
                fields.append(Field.create(my_field_name, comp, points,
                                            collection_size, timestep, TimeUnit.ISO_FORMAT))
                yac.def_couple(comp_name, grid_name, varname,
                               "plot", "plot_grid", my_field_name,
                               timestep, TimeUnit.ISO_FORMAT, 0, nnn)
del nnn

yac.enddef()

# create folders
for field in fields:
    for i in range(field.collection_size):
        os.makedirs(f"{field.name}_{i}", exist_ok=True)

def plot_buffer(field, buf):
    for i in range(buf.shape[0]):
        plt.imshow(buf[i,:].reshape((len(y),len(x)))[::-1,:])
        plt.title(f"{field.name} - {field.datetime}")
        plt.savefig(f"{field.name}_{i}/{field.datetime}.png")
    return buf

end_time = isodate.parse_datetime(yac.end_datetime)
buf = None
while isodate.parse_datetime(fields[0].datetime) <= end_time:
    field = fields[0]
    print(f"getting {field.name} at {field.datetime} ....")
    buf, info = field.get(buf)
    plot_buffer(field, buf)

    fields.sort(key=lambda f: isodate.parse_datetime(f.datetime))
